#include "Sculpt.h"

/** Constructor */
Sculpt::Sculpt() : mesh_(0), intensity_(0.75f), sculptMode_(INFLATE), topoMode_(UNIFORMISATION), centerPoint_(), culling_(false),
    detail_(0.75f), thickness_(0.5f), d2Min_(0.f), d2Max_(0.f), d2Thickness_(0.f), d2Move_(0.f), sweepCenter_(), sweepDir_()
{}

/** Destructor */
Sculpt::~Sculpt()
{}

/** Getters/Setters */
void Sculpt::setMesh(Mesh *mesh) { mesh_ = mesh; }
void Sculpt::setIntensity(float intensity) { intensity_ = intensity; }
void Sculpt::toggleCulling() { culling_ = !culling_; }
void Sculpt::setSculptMode(SculptMode mode) { sculptMode_ = mode; }
void Sculpt::setTopoMode(TopoMode mode) { topoMode_ = mode; }
void Sculpt::setDetail(float detail) { detail_ = detail; }
const QVector3D& Sculpt::getSweepCenter() { return sweepCenter_; }
void Sculpt::setSweepCenter(const QVector3D &center) { sweepCenter_ = center; }
void Sculpt::setSweepDir(const QVector3D &dir) { sweepDir_ = dir; }
bool Sculpt::isSweep() { return sculptMode_==SWEEP; }

/** Set sculpting mode */
void Sculpt::setAdaptiveParameters(float radiusSquared)
{
    d2Max_ = radiusSquared*(1.1f-detail_)*0.2f;
    d2Min_ = d2Max_/4.2025f;
    d2Move_ = d2Min_*0.2375f;
    d2Thickness_ = (4.f*d2Move_ + d2Max_/3.f)*1.1f;
}

/**
 * Sculpt the mesh. Main steps are :
 * 1. Take a bigger region topologically (2-ring vertices)
 * 2. Subdivide the triangles inside the brush radius (enlarged)
 * 3. Cull the vertices if asked (back face culling according to vertex normal)
 * 4. Select the vertices to be sculpted (not enlarged). Actually, old vertices to be sculpted
 *      are tagged during the picking process, whereas newly created vertices are tagged
 *      at the end of the subdivision process, so here, we just have to check the sculpt flag
 * 5. Sculpting operator
 * 6. Update the mesh (triangle's aabb, normals of triangle/vertices, octree and the VBO's)
*/
void Sculpt::sculptMesh(std::vector<int> &iVertsSelected, const QVector3D &intersectionPoint,
                        float radiusSquared, const QVector3D &eyeDirection)
{
    centerPoint_ = intersectionPoint;
    std::vector<Vertex> &vertices = mesh_->getVertices();

    std::vector<int> iTris = mesh_->getTrianglesFromVertices(iVertsSelected);

    //undo-redo
    mesh_->pushState(iTris,iVertsSelected);

    Topology topo(mesh_,radiusSquared,centerPoint_);
    setAdaptiveParameters(radiusSquared);
    switch(topoMode_)
    {
    case ADAPTIVE :
    case UNIFORMISATION : topo.uniformisation(iTris, d2Min_, d2Max_); break;
    case DECIMATION : topo.decimation(iTris, d2Min_); break;
    case SUBDIVISION : topo.subdivision(iTris, d2Max_); break;
    default : break;
    }

    iVertsSelected = mesh_->getVerticesFromTriangles(iTris);

    int nbVertsSelected = iVertsSelected.size();
    std::vector<int> iVertsSculpt;
    iVertsSculpt.reserve(nbVertsSelected);
    for(int i = 0; i<nbVertsSelected; ++i)
    {
        if(vertices[iVertsSelected[i]].sculptFlag_ == Vertex::sculptMask_)
            iVertsSculpt.push_back(iVertsSelected[i]);
    }

    std::vector<int> iVerts;
    int nbVertsSculpt = iVertsSculpt.size();
    if(culling_)
    {
        for (int i = 0; i<nbVertsSculpt; ++i)
        {
            if(QVector3D::dotProduct(eyeDirection, vertices[iVertsSculpt[i]].normal_)<=0)
                iVerts.push_back(iVertsSculpt[i]);
        }
    }
    else
        iVerts = iVertsSculpt;

    switch(sculptMode_)
    {
    case INFLATE :
    case DEFLATE : draw(iVerts, radiusSquared, intensity_); break;
    case SMOOTH : smooth(iVerts, intensity_); break;
    case FLATTEN : flatten(iVerts,radiusSquared, intensity_); break;
    case SWEEP : sweep(iVerts, radiusSquared, intensity_); break;
    default: break;
    }

    if(topoMode_==ADAPTIVE)
    {
        topo.adaptTopology(iTris, d2Thickness_);
        iVertsSelected = mesh_->getVerticesFromTriangles(iTris);
    }

    mesh_->updateMesh(iTris,iVertsSelected);
}

/** Smooth a group of vertices. New position is given by simple averaging */
void Sculpt::smooth(const std::vector<int> &iVerts, float intensity)
{
    std::vector<Vertex> &vertices = mesh_->getVertices();
    int nbVerts = iVerts.size();
    std::vector<QVector3D> smoothVerts(nbVerts);
    laplacianSmooth(iVerts,smoothVerts);
    if(topoMode_!=ADAPTIVE)
    {
#pragma omp parallel for
        for (int i = 0; i<nbVerts; ++i)
        {
            Vertex &vert = vertices[iVerts[i]];
            QVector3D displ = (smoothVerts[i]-vert)*intensity;
            vert+=displ;
        }
    }
    else
    {
        float dMove = sqrtf(d2Move_);
#pragma omp parallel for
        for (int i = 0; i<nbVerts; ++i)
        {
            Vertex &vert = vertices[iVerts[i]];
            QVector3D displ = (smoothVerts[i]-vert)*intensity;
            float displLength = displ.lengthSquared();
            if(displLength<=d2Move_)
                vert+=displ;
            else
                vert+=displ.normalized()*dMove;
        }
    }
}

/** Smooth a group of vertices along the plane defined by the normal of the vertex */
void Sculpt::smoothFlat(const std::vector<int> &iVerts, float intensity)
{
    std::vector<Vertex> &vertices = mesh_->getVertices();
    int nbVerts = iVerts.size();
    std::vector<QVector3D> smoothVerts(nbVerts);
    laplacianSmooth(iVerts,smoothVerts);
#pragma omp parallel for
    for (int i = 0; i<nbVerts; ++i)
    {
        Vertex &vert = vertices[iVerts[i]];
        QVector3D &vertSmo = smoothVerts[i];
        QVector3D &n = vert.normal_;
        vertSmo-=n*QVector3D::dotProduct(n,vertSmo-vert);
        vert+=(vertSmo-vert)*intensity;
    }
}

/**
 * Draw (inflate/deflate) a group of vertices. Move them according to the normal
 * representative of all the sculpting vertices. I couldn't come up with a good
 * falloff function, so it's just something that fall off strongly at the border of brush radius
*/
void Sculpt::draw(const std::vector<int> &iVerts, float radiusSquared, float intensity)
{
    std::vector<Vertex> &vertices = mesh_->getVertices();
    int nbVerts = iVerts.size();
    float radius = sqrtf(radiusSquared);
    float deformationIntensity = intensity*radius*0.1f;
    if(topoMode_==ADAPTIVE)
        deformationIntensity = std::min(sqrtf(d2Move_), deformationIntensity);
    if(sculptMode_==DEFLATE)
        deformationIntensity = -deformationIntensity;
#pragma omp parallel for
    for (int i = 0; i<nbVerts; ++i)
    {
        Vertex &vert=vertices[iVerts[i]];
        float dist = ((vert-centerPoint_).length()/radius);
        float fallOff = 3.f*dist*dist*dist*dist-4.f*dist*dist*dist+1.f;
        vert+=vert.normal_*deformationIntensity*fallOff;
    }
}

/** Compute average normal of a group of vertices with culling */
QVector3D Sculpt::areaNormal(const std::vector<int> &iVerts)
{
    std::vector<Vertex> &vertices = mesh_->getVertices();
    int nbVerts = iVerts.size();
    float areaX=0.f,areaY=0.f,areaZ=0.f;
#pragma omp parallel for reduction(+:areaX,areaY,areaZ)
    for (int i = 0; i<nbVerts; ++i)
    {
        const QVector3D &normal = vertices[iVerts[i]].normal_;
        areaX+=normal.x();
        areaY+=normal.y();
        areaZ+=normal.z();
    }
    return QVector3D(areaX,areaY,areaZ).normalized();
}

/** Compute average center of a group of vertices with culling */
QVector3D Sculpt::areaCenter(const std::vector<int> &iVerts)
{
    std::vector<Vertex> &vertices = mesh_->getVertices();
    int nbVerts = iVerts.size();
    float areaX=0.f,areaY=0.f,areaZ=0.f;
#pragma omp parallel for reduction(+:areaX,areaY,areaZ)
    for (int i = 0; i<nbVerts; ++i)
    {
        Vertex &v = vertices[iVerts[i]];
        areaX+=v.x();
        areaY+=v.y();
        areaZ+=v.z();
    }
    return QVector3D(areaX,areaY,areaZ)/nbVerts ;
}

/**
 * Flattening, projection of the sculpting vertex onto a plane
 * defined by the barycenter and normals of all the sculpting vertices
*/
void Sculpt::flatten(const std::vector<int> &iVerts, float radiusSquared, float intensity)
{
    QVector3D areaNorm = areaNormal(iVerts).normalized();
    QVector3D areaPoint = areaCenter(iVerts);
    if(areaNorm.lengthSquared()<0.0001f)
        return;
    std::vector<Vertex> &vertices = mesh_->getVertices();
    float radius = sqrtf(radiusSquared);
    int nbVerts = iVerts.size();
    float deformationIntensity = intensity * 0.3f;
    if(topoMode_!=ADAPTIVE)
    {
#pragma omp parallel for
        for (int i = 0; i<nbVerts; ++i)
        {
            Vertex &v = vertices[iVerts[i]];
            float distance = v.distanceToPlane(areaPoint,areaNorm);
            float dist = ((v-centerPoint_).length()/radius);
            float fallOff = 3.f*dist*dist*dist*dist-4.f*dist*dist*dist+1.f;
            v-=areaNorm*distance*deformationIntensity*fallOff;
        }
    }
    else
    {
        float dMove = sqrtf(d2Move_);
#pragma omp parallel for
        for (int i = 0; i<nbVerts; ++i)
        {
            Vertex &v = vertices[iVerts[i]];
            float distance = v.distanceToPlane(areaPoint,areaNorm);
            float dist = ((v-centerPoint_).length()/radius);
            float fallOff = 3.f*dist*dist*dist*dist-4.f*dist*dist*dist+1.f;
            v-=areaNorm*std::min(dMove,distance*deformationIntensity*fallOff);
        }
    }
}

/** Laplacian smooth. Special rule for vertex on the edge of the mesh. */
void Sculpt::laplacianSmooth(const std::vector<int> &iVerts, std::vector<QVector3D> &smoothVerts)
{
    std::vector<Vertex> &vertices = mesh_->getVertices();
    int nbVerts = iVerts.size();
#pragma omp parallel for
    for (int i = 0; i<nbVerts; ++i)
    {
        Vertex &vert = vertices[iVerts[i]];
        const std::vector<int> &ivRing = vert.ringVertices_;
        int nbVRing=ivRing.size();
        if(nbVRing!=(int)vert.tIndices_.size())
        {
            QVector3D center;
            int nbVertEdge = 0;
            for(int j = 0; j<nbVRing; ++j)
            {
                Vertex &ivr = vertices[ivRing[j]];
                if(ivr.tIndices_.size()!=ivr.ringVertices_.size())
                {
                    center+=ivr;
                    ++nbVertEdge;
                }
            }
            smoothVerts[i]=center/nbVertEdge ;
        }
        else
        {
            QVector3D center;
            for (int j=0;j<nbVRing;++j)
                center+=vertices[ivRing[j]];
            smoothVerts[i]=center/nbVRing;
        }
    }
}

/** Smooth a group of vertices along the plane defined by the normal of the vertex (no openMP) */
void Sculpt::smoothNoMp(const std::vector<int> &iVerts, bool flat)
{
    std::vector<Vertex> &vertices = mesh_->getVertices();
    int nbVerts = iVerts.size();
    std::vector<QVector3D> smoothVerts(nbVerts);
    for (int i = 0; i<nbVerts; ++i)
    {
        Vertex &vert = vertices[iVerts[i]];
        const std::vector<int> &ivRing = vert.ringVertices_;
        int nbVRing=ivRing.size();
        if(nbVRing!=(int)vert.tIndices_.size())
        {
            QVector3D center;
            int nbVertEdge = 0;
            for(int j=0;j<nbVRing;++j)
            {
                Vertex &ivr = vertices[ivRing[j]];
                if(ivr.tIndices_.size()!=ivr.ringVertices_.size())
                {
                    center+=ivr;
                    ++nbVertEdge;
                }
            }
            smoothVerts[i]=center/nbVertEdge ;
        }
        else
        {
            QVector3D center;
            for (int j=0;j<nbVRing;++j)
                center+=vertices[ivRing[j]];
            smoothVerts[i]=center/nbVRing;
        }
    }
    if(!flat)
    {
        for (int i = 0; i<nbVerts; ++i)
            vertices[iVerts[i]] = smoothVerts[i];
    }
    else
    {
        for (int i = 0; i<nbVerts; ++i)
        {
            Vertex &vert = vertices[iVerts[i]];
            QVector3D &vertSmo = smoothVerts[i];
            QVector3D &n = vert.normal_;
            vertSmo-=n*QVector3D::dotProduct(n,vertSmo-vert);
            vert = vertSmo;
        }
    }
}

/** Sweep deformation */
void Sculpt::sweep(const std::vector<int> &iVerts, float radiusSquared, float intensity)
{
    std::vector<Vertex> &vertices = mesh_->getVertices();
    int nbVerts = iVerts.size();
    float radius = sqrtf(radiusSquared);
    float deformationIntensity = intensity*radius*0.5f;

    if(topoMode_!=ADAPTIVE)
    {
#pragma omp parallel for
        for (int i = 0; i<nbVerts; ++i)
        {
            Vertex &vert = vertices[iVerts[i]];
            float dist = (vert-sweepCenter_).length()/radius;
            float fallOff = 3.f*dist*dist*dist*dist-4.f*dist*dist*dist+1.f;
            vert+=sweepDir_*deformationIntensity*fallOff;
        }
    }
    else
    {
#pragma omp parallel for
        for (int i = 0; i<nbVerts; ++i)
        {
            Vertex &vert = vertices[iVerts[i]];
            float dist = (vert-sweepCenter_).length()/radius;
            float fallOff = 3.f*dist*dist*dist*dist-4.f*dist*dist*dist+1.f;
            vert+=sweepDir_*std::min(d2Move_,deformationIntensity*fallOff);
        }
    }
}
