#ifndef __MESH_H__
#define __MESH_H__

#include <string>
#include <QtOpenGL>
#include <algorithm>
#include <stdint.h>
#include "Tools.h"
#include "Camera.h"
#include "Octree.h"
#include "Triangle.h"
#include "Vertex.h"
#include "State.h"
#include "Window.h"
#include "omp.h"

/**
 * Mesh
 * @author Stphane GINIER
 */
class Mesh
{
public:
    static const float globalScale_ = 500.f; //for precision issue...
    static int stateMask_; //for history

public:
    enum ShaderMode{PHONG, WIREFRAME, TRANSPARENCY, REFLECTION_1, REFLECTION_2};
    Mesh();
    ~Mesh();
    std::vector<Triangle>& getTriangles();
    std::vector<Vertex>& getVertices();
    std::vector<Octree*>& getLeavesUpdate();
    Triangle& getTriangle(int i);
    Vertex& getVertex(int i);
    int getNbTriangles() const;
    int getNbVertices() const;
    QVector3D getCenter() const;
    Octree* getOctree() const;
    QColor getColor() const;
    float getScale() const;
    void setColor(const QColor &color);
    void setIsSelected(bool);
    bool getDisplayOctree() const;
    void toggleDisplayOctree();
    QMatrix4x4& getTransformation();
    QMatrix4x4 getInverseTransformation() const;
    ShaderMode getShader() const;

    std::vector<int> getTrianglesFromVertices(const std::vector<int> &iVerts);
    std::vector<int> getVerticesFromTriangles(const std::vector<int> &iTris);
    void expandTriangles(std::vector<int> &iTris, int nRing);
    void expandVertices(std::vector<int> &iVerts, int nRing);
    void computeRingVertices(int iVert);

    void setShader(ShaderMode shader);
    QVector3D getTriangleCenter(int iTri) const;
    void moveTo(const QVector3D &destination);
    void setTransformation(QMatrix4x4 matTransform);

    void draw(const Camera &camera, const QVector3D &centerPicking = QVector3D(), float radiusSquared=0);
    void initVBO(Window *window = 0);
    void initMesh(Window *window = 0);

    void updateMesh(const std::vector<int> &iTris, const std::vector<int> &iVerts);

    std::vector<int> subdivide(std::vector<int> &iTris,std::vector<int> &iVerts,float inradiusMaxSquared);
    void triangleSubdivision(int iTri);

    void checkLeavesUpdate();

    //undo-redo
    std::vector<Triangle>& getTrianglesState();
    std::vector<Vertex>& getVerticesState();
    void startPushState();
    void pushState(const std::vector<int> &iTris, const std::vector<int> &iVerts);
    void undo();
    void redo();
    void recomputeOctree(const Aabb &aabbSplit);

private:
    void updateOctree(const std::vector<int> &iTris);
    void updateNormals(const std::vector<int> &iVerts);
    float angleTri(int iTri, int iVer);
    void updateVertexBuffer(const std::vector<int> &iVerts);
    void updateNormalBuffer(const std::vector<int> &iVerts);
    void updateIndexBuffer(const std::vector<int> &iTris);
    void updateTransformation();
    void drawBufferShader();
    std::vector<Vertex> vertices_; //vertices
    std::vector<Triangle> triangles_; //triangles
    QGLBuffer verticesBuffer_; //vertices buffer (openGL)
    QGLBuffer normalsBuffer_; //normals buffer (openGL)
    QGLBuffer indicesBuffer_; //indexes (openGL)
    QVector3D center_; //center of mesh
    ShaderMode shaderType_; //type of shader
    QGLShaderProgram shaderProgram_; //program shader
    QColor color_; //main color
    float scale_; //scale
    Octree *octree_; //octree
    bool displayOctree_; //if the octree is displayed
    QMatrix4x4 matTransform_; //transformation matrix of the mesh
    GLfloat matTransformArray_[16]; //transformation matrix of the mesh (openGL)
    GLuint reflectionTexture_; //texture reflection
    std::vector<Octree*> leavesUpdate_; //leaves of the octree to check

    //undo-redo
    std::list<State> undo_; //undo actions
    std::list<State> redo_; //redo actions
    std::list<State>::iterator undoIte_; //iterator to undo
    bool beginIte_; //end of undo action
};

#endif /*__MESH_H__*/
